@@ -1,14 +1,7 @@
-import os
 import sys
-import logging
-import pkgutil
-import importlib.machinery
-import importlib.util
-import types
-import shutil
-from copy import deepcopy
-import traceback
+import os
 import argparse
+import logging
 import requests
 import jsonpickle
 from pathlib import Path
@@ -16,6 +9,13 @@ from tqdm import tqdm
 from zipfile import ZipFile
 from distutils.dir_util import mkpath
 import operator
+import pkgutil
+import importlib.machinery
+import importlib.util
+import types
+import shutil
+from copy import deepcopy
+import traceback
 import inspect
 import re
 from subprocess import Popen
@@ -23,6 +23,9 @@ from subprocess import PIPE
 from subprocess import STDOUT
 
 ######## START CONTENT ########
+def INVALID_NAME():
+	return "INVALID_NAME"
+
 class ActualType(type):
 	def __repr__(self):
 		return self.__name__
@@ -125,9 +128,6 @@ class SelfRegistering(object):
 		if (directory not in sys.path):
 			sys.path.append(directory)
 
-def INVALID_NAME():
-	return "INVALID_NAME"
-
 
 # A Datum is a base class for any object-oriented class structure.
 # This class is intended to be derived from and added to.
@@ -165,6 +165,165 @@ class Datum(SelfRegistering):
 		this.valid = False
 
 
+# A DataContainer allows Data to be stored and worked with.
+# This class is intended to be derived from and added to.
+# Each DataContainer is comprised of multiple Data (see Datum.py for more).
+# NOTE: DataContainers are, themselves Data. Thus, you can nest your child classes however you would like.
+class DataContainer(Datum):
+
+	def __init__(this, name=INVALID_NAME()):
+		super().__init__(name)
+
+		# The data *this contains.
+		this.data = []
+
+
+	# RETURNS: an empty, invalid Datum.
+	def InvalidDatum(this):
+		ret = Datum()
+		ret.Invalidate()
+		return ret
+
+
+	# Sort things! Requires by be a valid attribute of all Data.
+	def SortData(this, by):
+		this.data.sort(key=operator.attrgetter(by))
+
+
+	# Adds a Datum to *this
+	def AddDatum(this, datum):
+		this.data.append(datum)
+
+
+	# RETURNS: a Datum with datumAttribute equal to match, an invalid Datum if none found.
+	def GetDatumBy(this, datumAttribute, match):
+		for d in this.data:
+			try: # within for loop 'cause maybe there's an issue with only 1 Datum and the rest are fine.
+				if (str(getattr(d, datumAttribute)) == str(match)):
+					return d
+			except Exception as e:
+				logging.error(f"{this.name} - {e.message}")
+				continue
+		return this.InvalidDatum()
+
+
+	# RETURNS: a Datum of the given name, an invalid Datum if none found.
+	def GetDatum(this, name):
+		return this.GetDatumBy('name', name)
+
+
+	# Removes all Data in toRem from *this.
+	# RETURNS: the Data removed
+	def RemoveData(this, toRem):
+		# logging.debug(f"Removing {toRem}")
+		this.data = [d for d in this.data if d not in toRem]
+		return toRem
+
+
+	# Removes all Data which match toRem along the given attribute
+	def RemoveDataBy(this, datumAttribute, toRem):
+		toRem = [d for d in this.data if str(getattr(d, datumAttribute)) in list(map(str, toRem))]
+		return this.RemoveData(toRem)
+
+
+	# Removes all Data in *this except toKeep.
+	# RETURNS: the Data removed
+	def KeepOnlyData(this, toKeep):
+		toRem = [d for d in this.data if d not in toKeep]
+		return this.RemoveData(toRem)
+
+
+	# Removes all Data except those that match toKeep along the given attribute
+	# RETURNS: the Data removed
+	def KeepOnlyDataBy(this, datumAttribute, toKeep):
+		# logging.debug(f"Keeping only class with a {datumAttribute} of {toKeep}")
+		# toRem = []
+		# for d in this.class:
+		#	 shouldRem = False
+		#	 for k in toKeep:
+		#		 if (str(getattr(d, datumAttribute)) == str(k)):
+		#			 logging.debug(f"found {k} in {d.__dict__}")
+		#			 shouldRem = True
+		#			 break
+		#	 if (shouldRem):
+		#		 toRem.append(d)
+		#	 else:
+		#		 logging.debug(f"{k} not found in {d.__dict__}")
+		toRem = [d for d in this.data if str(getattr(d, datumAttribute)) not in list(map(str, toKeep))]
+		return this.RemoveData(toRem)
+
+
+	# Removes all Data with the name "INVALID NAME"
+	# RETURNS: the removed Data
+	def RemoveAllUnlabeledData(this):
+		toRem = []
+		for d in this.data:
+			if (d.name =="INVALID NAME"):
+				toRem.append(d)
+		return this.RemoveData(toRem)
+
+
+	# Removes all invalid Data
+	# RETURNS: the removed Data
+	def RemoveAllInvalidData(this):
+		toRem = []
+		for d in this.data:
+			if (not d.IsValid()):
+				toRem.append(d)
+		return this.RemoveData(toRem)
+
+
+	# Removes all Data that have an attribute value relative to target.
+	# The given relation can be things like operator.le (i.e. <=)
+	#   See https://docs.python.org/3/library/operator.html for more info.
+	# If ignoreNames is specified, any Data of those names will be ignored.
+	# RETURNS: the Data removed
+	def RemoveDataRelativeToTarget(this, datumAttribute, relation, target, ignoreNames = []):
+		try:
+			toRem = []
+			for d in this.data:
+				if (ignoreNames and d.name in ignoreNames):
+					continue
+				if (relation(getattr(d, datumAttribute), target)):
+					toRem.append(d)
+			return this.RemoveData(toRem)
+		except Exception as e:
+			logging.error(f"{this.name} - {e.message}")
+			return []
+
+
+	# Removes any Data that have the same datumAttribute as a previous Datum, keeping only the first.
+	# RETURNS: The Data removed
+	def RemoveDuplicateDataOf(this, datumAttribute):
+		toRem = [] # list of Data
+		alreadyProcessed = [] # list of strings, not whatever datumAttribute is.
+		for d1 in this.data:
+			skip = False
+			for dp in alreadyProcessed:
+				if (str(getattr(d1, datumAttribute)) == dp):
+					skip = True
+					break
+			if (skip):
+				continue
+			for d2 in this.data:
+				if (d1 is not d2 and str(getattr(d1, datumAttribute)) == str(getattr(d2, datumAttribute))):
+					logging.info(f"Removing duplicate Datum {d2} with unique id {getattr(d2, datumAttribute)}")
+					toRem.append(d2)
+					alreadyProcessed.append(str(getattr(d1, datumAttribute)))
+		return this.RemoveData(toRem)
+
+
+	# Adds all Data from otherDataContainer to *this.
+	# If there are duplicate Data identified by the attribute preventDuplicatesOf, they are removed.
+	# RETURNS: the Data removed, if any.
+	def ImportDataFrom(this, otherDataContainer, preventDuplicatesOf=None):
+		this.data.extend(otherDataContainer.data);
+		if (preventDuplicatesOf is not None):
+			return this.RemoveDuplicateDataOf(preventDuplicatesOf)
+		return []
+
+
+
 # util is a namespace for any miscellaneous utilities.
 # You cannot create a util.
 class util:
@@ -271,6 +430,35 @@ class util:
 
 		resetStyle = "\033[0m"
 
+
+	# Add a logging level
+	# per: https://stackoverflow.com/questions/2183233/how-to-add-a-custom-loglevel-to-pythons-logging-facility/35804945#35804945
+	@staticmethod
+	def AddLoggingLevel(level, value):
+		levelName = level.upper()
+		methodName = level.lower()
+
+		if hasattr(logging, levelName):
+			raise AttributeError('{} already defined in logging module'.format(levelName))
+		if hasattr(logging, methodName):
+			raise AttributeError('{} already defined in logging module'.format(methodName))
+		if hasattr(logging.getLogger(), methodName):
+			raise AttributeError('{} already defined in logger class'.format(methodName))
+
+		# This method was inspired by the answers to Stack Overflow post
+		# http://stackoverflow.com/q/2183233/2988730, especially
+		# http://stackoverflow.com/a/13638084/2988730
+		def logForLevel(this, message, *args, **kwargs):
+			if this.isEnabledFor(value):
+				this._log(value, message, args, **kwargs)
+		def logToRoot(message, *args, **kwargs):
+			logging.log(value, message, *args, **kwargs)
+
+		logging.addLevelName(value, levelName)
+		setattr(logging, levelName, value)
+		setattr(logging.getLogger(), methodName, logForLevel)
+		setattr(logging, methodName, logToRoot)
+
 # Don't import Method or Executor, even though they are required: it will cause a circular dependency.
 # Instead, pretend there's a forward declaration here and don't think too hard about it ;)
 ################################################################################
@@ -442,14 +630,14 @@ class Functor(Datum):
 
 		if (isinstance(value, dict)):
 			ret = {}
-			for key, value in value.items():
-				ret[key] = this.EvaluateToType(value)
+			for key, val in value.items():
+				ret[key] = this.EvaluateToType(val, evaluateExpressions)
 			return ret
 
 		if (isinstance(value, list)):
 			ret = []
-			for value in value:
-				ret.append(this.EvaluateToType(value))
+			for val in value:
+				ret.append(this.EvaluateToType(val, evaluateExpressions))
 			return ret
 
 		if (isinstance(value, str)):
@@ -638,6 +826,7 @@ class Functor(Datum):
 						continue
 
 					methodToInsert = deepcopy(method)
+					methodToInsert.UpdateSource()
 
 					if (existingMethod.inheritedMethodsFirst):
 						logging.debug(f"Will call {method.name} from {source} to prior to this.")
@@ -648,6 +837,7 @@ class Functor(Datum):
 						this.methods[method.name].next.append(methodToInsert)
 				else:
 					this.methods[method.name] = deepcopy(method)
+					this.methods[method.name].UpdateSource()
 
 		for method in this.methods.values():
 			logging.debug(f"Populating method {this.name}.{method.name}({', '.join([a for a in method.requiredKWArgs] + [a+'='+str(v) for a,v in method.optionalKWArgs.items()])})")
@@ -716,7 +906,8 @@ class Functor(Datum):
 			'environment',
 			'executor'
 		]
-		this.Set('next', this.FetchWithout(dontFetchFrom, 'next', []))
+		# Let 'next' evaluate its expressions if it chooses to. We don't need to do that pre-emptively.
+		this.Set('next', this.FetchWithout(dontFetchFrom, 'next', []), evaluateExpressions=False)
 
 
 	# Make sure that precursors have provided all necessary methods for *this.
@@ -773,8 +964,8 @@ class Functor(Datum):
 			this.PopulateMethods() # Doesn't require Fetch; depends on precursor
 			this.ParseInitialArgs() # Usually where config is read in.
 			this.ValidateStaticArgs() # nop on call 2+
-			this.ValidateArgs()
 			this.PopulateNext()
+			this.ValidateArgs()
 			this.ValidateMethods()
 
 			this.BeforeFunction()
@@ -824,6 +1015,24 @@ class Functor(Datum):
 		return this.executor
 
 
+	# Add support for deepcopy.
+	# Copies everything besides methods; those will be created by PopulateMethods or removed.
+	def __deepcopy__(this, memodict=None):
+		logging.debug(f"Creating new {this.__class__} from {this.name}")
+		cls = this.__class__
+		ret = cls.__new__(cls)
+		ret.__init__()
+
+		memodict[id(this)] = ret
+		for key, val in [tup for tup in this.__dict__.items() if tup[0] not in ['methods']]:
+			if (callable(val)):
+				# PopulateMethods will take care of recreating skipped Methods
+				# All other methods are dropped since they apparently have problems on some implementations.
+				continue
+			setattr(ret, key, deepcopy(val, memodict))
+		return ret
+
+
 	######## START: Fetch Locations ########
 
 	def fetch_location_this(this, varName, default, fetchFrom, attempted):
@@ -874,165 +1083,6 @@ class Functor(Datum):
 
 	######## END: Fetch Locations ########
 
-
-# A DataContainer allows Data to be stored and worked with.
-# This class is intended to be derived from and added to.
-# Each DataContainer is comprised of multiple Data (see Datum.py for more).
-# NOTE: DataContainers are, themselves Data. Thus, you can nest your child classes however you would like.
-class DataContainer(Datum):
-
-	def __init__(this, name=INVALID_NAME()):
-		super().__init__(name)
-
-		# The data *this contains.
-		this.data = []
-
-
-	# RETURNS: an empty, invalid Datum.
-	def InvalidDatum(this):
-		ret = Datum()
-		ret.Invalidate()
-		return ret
-
-
-	# Sort things! Requires by be a valid attribute of all Data.
-	def SortData(this, by):
-		this.data.sort(key=operator.attrgetter(by))
-
-
-	# Adds a Datum to *this
-	def AddDatum(this, datum):
-		this.data.append(datum)
-
-
-	# RETURNS: a Datum with datumAttribute equal to match, an invalid Datum if none found.
-	def GetDatumBy(this, datumAttribute, match):
-		for d in this.data:
-			try: # within for loop 'cause maybe there's an issue with only 1 Datum and the rest are fine.
-				if (str(getattr(d, datumAttribute)) == str(match)):
-					return d
-			except Exception as e:
-				logging.error(f"{this.name} - {e.message}")
-				continue
-		return this.InvalidDatum()
-
-
-	# RETURNS: a Datum of the given name, an invalid Datum if none found.
-	def GetDatum(this, name):
-		return this.GetDatumBy('name', name)
-
-
-	# Removes all Data in toRem from *this.
-	# RETURNS: the Data removed
-	def RemoveData(this, toRem):
-		# logging.debug(f"Removing {toRem}")
-		this.data = [d for d in this.data if d not in toRem]
-		return toRem
-
-
-	# Removes all Data which match toRem along the given attribute
-	def RemoveDataBy(this, datumAttribute, toRem):
-		toRem = [d for d in this.data if str(getattr(d, datumAttribute)) in list(map(str, toRem))]
-		return this.RemoveData(toRem)
-
-
-	# Removes all Data in *this except toKeep.
-	# RETURNS: the Data removed
-	def KeepOnlyData(this, toKeep):
-		toRem = [d for d in this.data if d not in toKeep]
-		return this.RemoveData(toRem)
-
-
-	# Removes all Data except those that match toKeep along the given attribute
-	# RETURNS: the Data removed
-	def KeepOnlyDataBy(this, datumAttribute, toKeep):
-		# logging.debug(f"Keeping only class with a {datumAttribute} of {toKeep}")
-		# toRem = []
-		# for d in this.class:
-		#	 shouldRem = False
-		#	 for k in toKeep:
-		#		 if (str(getattr(d, datumAttribute)) == str(k)):
-		#			 logging.debug(f"found {k} in {d.__dict__}")
-		#			 shouldRem = True
-		#			 break
-		#	 if (shouldRem):
-		#		 toRem.append(d)
-		#	 else:
-		#		 logging.debug(f"{k} not found in {d.__dict__}")
-		toRem = [d for d in this.data if str(getattr(d, datumAttribute)) not in list(map(str, toKeep))]
-		return this.RemoveData(toRem)
-
-
-	# Removes all Data with the name "INVALID NAME"
-	# RETURNS: the removed Data
-	def RemoveAllUnlabeledData(this):
-		toRem = []
-		for d in this.data:
-			if (d.name =="INVALID NAME"):
-				toRem.append(d)
-		return this.RemoveData(toRem)
-
-
-	# Removes all invalid Data
-	# RETURNS: the removed Data
-	def RemoveAllInvalidData(this):
-		toRem = []
-		for d in this.data:
-			if (not d.IsValid()):
-				toRem.append(d)
-		return this.RemoveData(toRem)
-
-
-	# Removes all Data that have an attribute value relative to target.
-	# The given relation can be things like operator.le (i.e. <=)
-	#   See https://docs.python.org/3/library/operator.html for more info.
-	# If ignoreNames is specified, any Data of those names will be ignored.
-	# RETURNS: the Data removed
-	def RemoveDataRelativeToTarget(this, datumAttribute, relation, target, ignoreNames = []):
-		try:
-			toRem = []
-			for d in this.data:
-				if (ignoreNames and d.name in ignoreNames):
-					continue
-				if (relation(getattr(d, datumAttribute), target)):
-					toRem.append(d)
-			return this.RemoveData(toRem)
-		except Exception as e:
-			logging.error(f"{this.name} - {e.message}")
-			return []
-
-
-	# Removes any Data that have the same datumAttribute as a previous Datum, keeping only the first.
-	# RETURNS: The Data removed
-	def RemoveDuplicateDataOf(this, datumAttribute):
-		toRem = [] # list of Data
-		alreadyProcessed = [] # list of strings, not whatever datumAttribute is.
-		for d1 in this.data:
-			skip = False
-			for dp in alreadyProcessed:
-				if (str(getattr(d1, datumAttribute)) == dp):
-					skip = True
-					break
-			if (skip):
-				continue
-			for d2 in this.data:
-				if (d1 is not d2 and str(getattr(d1, datumAttribute)) == str(getattr(d2, datumAttribute))):
-					logging.info(f"Removing duplicate Datum {d2} with unique id {getattr(d2, datumAttribute)}")
-					toRem.append(d2)
-					alreadyProcessed.append(str(getattr(d1, datumAttribute)))
-		return this.RemoveData(toRem)
-
-
-	# Adds all Data from otherDataContainer to *this.
-	# If there are duplicate Data identified by the attribute preventDuplicatesOf, they are removed.
-	# RETURNS: the Data removed, if any.
-	def ImportDataFrom(this, otherDataContainer, preventDuplicatesOf=None):
-		this.data.extend(otherDataContainer.data);
-		if (preventDuplicatesOf is not None):
-			return this.RemoveDuplicateDataOf(preventDuplicatesOf)
-		return []
-
-
 #from .Executor import Executor # don't import this, it'll be circular!
 
 # @recoverable
@@ -1081,7 +1131,7 @@ def RecoverableImplementation(obj, executor, function, *args, **kwargs):
 				logging.debug(f"Trying function ({function}) again after applying {res}.")
 				ret = function(obj, *args, **kwargs)
 				executor.ClearErrorResolutionStack(str(e)) # success!
-				logging.info(f"{res} successfully resolved '{e}'!")
+				logging.recovery(f"{res} successfully resolved '{e}'!")
 				logging.debug(f"Error stack is now: {executor.errorResolutionStack}")
 				return ret
 			except Exception as e2:
@@ -1197,7 +1247,15 @@ class Executor(DataContainer, Functor):
 
 	# Global logging config.
 	# Override this method to disable or change.
+	# This method will add a 'setupBy' member to the root logger in order to ensure no one else (e.g. another Executor) tries to reconfigure the logger while we're using it.
+	# The 'setupBy' member will be removed from the root logger by TeardownLogging, which is called in AfterFunction().
 	def SetupLogging(this):
+		try:
+			util.AddLoggingLevel('recovery', logging.ERROR+1)
+		except:
+			# Could already be setup.
+			pass
+
 		class CustomFormatter(logging.Formatter):
 
 			preFormat = util.console.GetColorCode('white', 'dark') + '%(asctime)s'
@@ -1209,6 +1267,7 @@ class Executor(DataContainer, Functor):
 				logging.INFO: preFormat + util.console.GetColorCode('white', 'light') + format + postFormat,
 				logging.WARNING: preFormat + util.console.GetColorCode('yellow', 'light') + format + postFormat,
 				logging.ERROR: preFormat + util.console.GetColorCode('red', 'dark') + format + postFormat,
+				logging.RECOVERY: preFormat + util.console.GetColorCode('green', 'light') + format + postFormat,
 				logging.CRITICAL: preFormat + util.console.GetColorCode('red', 'light', styles=['bold']) + format + postFormat
 			}
 
@@ -1217,16 +1276,32 @@ class Executor(DataContainer, Functor):
 				formatter = logging.Formatter(log_fmt, datefmt = '%H:%M:%S')
 				return formatter.format(record)
 
+		# Skip setting up logging if it's already been done.
+		if (hasattr(logging.getLogger(), 'setupBy')):
+			return
+
 		logging.getLogger().handlers.clear()
 		stderrHandler = logging.StreamHandler()
 		stderrHandler.setLevel(logging.CRITICAL)
 		stderrHandler.setFormatter(CustomFormatter())
 		logging.getLogger().addHandler(stderrHandler)
+		setattr(logging.getLogger(), 'setupBy', this)
+
+
+	# Global logging de-config.
+	def TeardownLogging(this):
+		if (not hasattr(logging.getLogger(), 'setupBy')):
+			return
+		if (not logging.getLogger().setupBy == this):
+			return
+		delattr(logging.getLogger(), 'setupBy')
+
 
 
 	# Logging to stderr is easy, since it will always happen.
 	# However, we also want the user to be able to log to a file, possibly set in the config.json, which requires a Fetch().
 	# Thus, setting up the log file handler has to occur later than the initial SetupLogging call.
+	# Calling this multiple times will add multiple log handlers.
 	def SetLogFile(this):
 		this.Set('log_file', this.Fetch('log_file', None, ['args', 'config', 'environment']))
 		if (this.log_file is None):
@@ -1324,21 +1399,22 @@ class Executor(DataContainer, Functor):
 
 	# How do we get the verbosity level and what do we do with it?
 	# This method should set log levels, etc.
-	def SetVerbosity(this):
+	def SetVerbosity(this, fetch=True):
 
-		# Take the highest of -v vs --verbosity
-		verbosity = this.Fetch('verbosity', 0, ['args', 'config', 'environment'])
-		if (verbosity > this.verbosity):
-			logging.debug(f"Setting verbosity to {verbosity}") # debug statements will be available when using external systems, like pytest.
-			this.verbosity = verbosity
+		if (fetch):
+			# Take the highest of -v vs --verbosity
+			verbosity = this.Fetch('verbosity', 0, ['args', 'config', 'environment'])
+			if (verbosity > this.verbosity):
+				logging.debug(f"Setting verbosity to {verbosity}") # debug statements will be available when using external systems, like pytest.
+				this.verbosity = verbosity
 
-		if (this.parsedArgs.verbose == 1):
+		if (this.verbosity == 1):
 			logging.getLogger().handlers[0].setLevel(logging.WARNING)
 			logging.getLogger().setLevel(logging.WARNING)
-		elif (this.parsedArgs.verbose == 2):
+		elif (this.verbosity == 2):
 			logging.getLogger().handlers[0].setLevel(logging.INFO)
 			logging.getLogger().setLevel(logging.INFO)
-		elif(this.parsedArgs.verbose >= 3):
+		elif(this.verbosity >= 3):
 			logging.getLogger().handlers[0].setLevel(logging.DEBUG)
 			logging.getLogger().setLevel(logging.DEBUG)
 
@@ -1351,6 +1427,9 @@ class Executor(DataContainer, Functor):
 
 		this.verbosity = this.parsedArgs.verbose
 
+		# If verbosity was specified on the command line, let's print more info while reading in the config, etc.
+		this.SetVerbosity(False)
+
 		extraArgsKeys = []
 		for index in range(0, len(extraArgs), 2):
 			keyStr = extraArgs[index]
@@ -1388,6 +1467,11 @@ class Executor(DataContainer, Functor):
 		pass
 
 
+	# Close out anything we left open.
+	def AfterFunction(this):
+		this.TeardownLogging()
+
+
 	# Execute a Functor based on name alone (not object).
 	# If the given Functor has been Executed before, the cached Functor will be called again. Otherwise, a new Functor will be constructed.
 	@recoverable
